﻿Module pngchunk

    Sub Main()
        If My.Application.CommandLineArgs.Count = 0 Then
            Console.WriteLine("pngchunk - http://long-cat.net")
            Console.WriteLine("Usage: pngchunk [options] file.png")
            Console.WriteLine()
            Console.WriteLine("Options:" & vbNewLine &
                              "-x   extract embedded PNG files and non-PNG data")
            Exit Sub
        End If

        Dim extract As Boolean
        Dim filename As String = String.Empty
        Dim pngcount, datcount As Integer

        For Each opt In My.Application.CommandLineArgs
            If opt = "-x" Or opt = "/x" Then
                extract = True
            Else
                filename = opt
            End If
        Next

        If Not FileIO.FileSystem.FileExists(filename) Then
            Console.WriteLine("File {0} not found!", filename)
        End If
        Dim fs As New IO.FileStream(filename, IO.FileMode.Open, IO.FileAccess.Read, IO.FileShare.Read)
        Dim fi As New IO.FileInfo(filename)

        While Not fs.Position = fs.Length
            Dim startposition As Long = fs.Position
            FindHeader(fs)

            If fs.Position - 8 <> startposition And fs.Position <> fs.Length Then
                Console.WriteLine(">> Skipped {0} bytes of non-png data", fs.Position - 8 - startposition)
                If extract Then
                    CopyPart(filename, fi.FullName & "." & datcount & ".dat", startposition, fs.Position - 8 - startposition)
                    datcount += 1
                End If

            ElseIf fs.Position = fs.Length Then
                Console.WriteLine(">> Skipped {0} bytes of non-png data", fs.Position - startposition)
                If extract Then
                    CopyPart(filename, fi.FullName & "." & datcount & ".dat", startposition, fs.Position - startposition)
                    datcount += 1
                End If

            End If

            startposition = fs.Position - 8

            If fs.Position = fs.Length Then
                Console.WriteLine("End of file")
                fs.Close()
                Exit Sub
            Else
                Console.WriteLine("-- PNG header at {0}", fs.Position - 8)
            End If


            Dim buf() As Byte
            Dim len As Integer

            Dim chunklength As UInt32
            Dim chunktype As String = String.Empty

            While chunktype <> "IEND" AndAlso fs.Position <> fs.Length
                ReDim buf(3)
                len = fs.Read(buf, 0, 4)
                If len <> 4 Then
                    Console.WriteLine("Incomplete chunk length header")
                    Exit While
                End If

                'big endian
                chunklength = &H1000000L * buf(0) + &H10000L * buf(1) + &H100L * buf(2) + buf(3)
                Console.Write("{0,10}:", chunklength)

                ReDim buf(chunklength + 3)
                len = fs.Read(buf, 0, chunklength + 4)
                If len <> chunklength + 4 Then
                    Console.WriteLine("Incomplete chunk data")
                    Exit While
                End If
                chunktype = System.Text.Encoding.ASCII.GetString(buf, 0, 4)
                Console.Write(chunktype & " ")
                Dim datachecksum As UInt32 = Crc32.ComputeChecksum(buf)

                ReDim buf(3)
                len = fs.Read(buf, 0, 4)
                If len <> 4 Then
                    Console.WriteLine("Incomplete CRC data")
                End If
                Dim chunkchecksum As UInt32 = &H1000000L * buf(0) + &H10000L * buf(1) + &H100L * buf(2) + buf(3)
                If chunkchecksum = datachecksum Then
                    Console.Write("ok")
                Else
                    Console.Write("INVALID CRC")
                End If
                Console.WriteLine()

            End While

            If extract Then
                CopyPart(filename, fi.FullName & "." & pngcount & ".png", startposition, fs.Position - startposition)
                pngcount += 1
            End If

        End While

        fs.Close()
    End Sub

    Public Sub FindHeader(fs As IO.FileStream)
        Dim header() As Byte = {&H89, &H50, &H4E, &H47, &HD, &HA, &H1A, &HA}
        Dim i As Integer = 0
        Dim startposition As Long = fs.Position
        While i < 8
            If fs.ReadByte = header(i) Then
                i += 1
            Else
                i = 0
            End If
            If fs.Position = fs.Length Then
                Exit Sub
            End If
        End While
    End Sub

    Public Sub CopyPart(srcfile As String, destfile As String, offset As Long, bytes As Long)
        Dim srcfs As New IO.FileStream(srcfile, IO.FileMode.Open, IO.FileAccess.Read, IO.FileShare.Read)
        Dim destfs As New IO.FileStream(destfile, IO.FileMode.Create, IO.FileAccess.Write, IO.FileShare.Read)
        srcfs.Seek(offset, IO.SeekOrigin.Begin)
        Dim buf(bytes) As Byte
        Dim len As Integer = srcfs.Read(buf, 0, bytes)
        destfs.Write(buf, 0, len)
        srcfs.Close()
        destfs.Close()
    End Sub

    Public Class Crc32
        Shared table As UInteger()

        Shared Sub New()
            Dim poly As UInteger = &HEDB88320UI
            table = New UInteger(255) {}
            Dim temp As UInteger = 0
            For i As UInteger = 0 To table.Length - 1
                temp = i
                For j As Integer = 8 To 1 Step -1
                    If (temp And 1) = 1 Then
                        temp = CUInt((temp >> 1) Xor poly)
                    Else
                        temp >>= 1
                    End If
                Next
                table(i) = temp
            Next
        End Sub

        Public Shared Function ComputeChecksum(bytes As Byte()) As UInteger
            Dim crc As UInteger = &HFFFFFFFFUI
            For i As Integer = 0 To bytes.Length - 1
                Dim index As Byte = CByte(((crc) And &HFF) Xor bytes(i))
                crc = CUInt((crc >> 8) Xor table(index))
            Next
            Return Not crc
        End Function
    End Class

End Module
